/*
bingo Version 0.1.0 - Bingo for Curses
Copyright (C) 2005  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ncurses.h>
#include <time.h>
#include <sys/times.h>
#include <errno.h>
#include "rnd.h"

typedef struct glbfmt {
   int run;          /* 0 = quit, 1 = run         */
   int reset;        /* 0 = no reset, 1 = reset   */
   unsigned char *sd;     /* random number seed   */
   int drawnum;      /* number drawn from urn     */
   int prevnum;      /* previous number drawn     */
   int hit;          /* 0 = no match, 1 = match   */
   int bingo;        /* 0 = no bingo, 1 = bingo   */
   char loc[128];    /* location of bingo         */
   int wait;         /* wait time in seconds      */
   int log;          /* handle for log file       */
   int card[25];     /* bingo card                */
   int cover[25];    /* covered squares           */
   int urn[75];      /* 1 - 75 in random sequence */
   } glbfmt;

void putstx(char *pgm)
   {
   fprintf(stderr,"Usage: %s interval\n", pgm);
   fprintf(stderr,"Where interval is 1-10 seconds\n");
   fprintf(stderr,"Example: %s 5\n", pgm);
   exit(1);
   } /* putstx */

void opn(glbfmt *glb)
   {
   glb->log = open("bingo.log",
      O_WRONLY|O_CREAT|O_TRUNC,
      S_IRUSR|S_IWUSR);
   if (glb->log < 0)
      {
      fprintf(stderr,"bingo.log\n");
      perror("Open error");
      exit(1);
      } /* if open error */
   } /* opn */

void putlog(glbfmt *glb, char *str)
   {
   int len;
   int wrtlen;
   len = strlen(str);
   wrtlen = write(glb->log,str,len);
   if (wrtlen != len)
      {
      fprintf(stderr,"bingo.log\n");
      perror("Write error");
      exit(1);
      } /* if write error */
   } /* putlog */

void cls(glbfmt *glb)
   {
   int rslt;
   rslt = close(glb->log);
   if (rslt < 0)
      {
      fprintf(stderr,"bingo.log\n");
      perror("Close error");
      exit(1);
      } /* if close error */
   } /* cls */

void putcard(glbfmt *glb)
   {
   int i,j;
   int *p;
   time_t now;
   struct tm *t;
   char *tstr;
   char cell[16];
   char str[128];
   time(&now);
   t = (struct tm *) localtime(&now);
   tstr = (char *) asctime(t);
   putlog(glb,tstr);
   putlog(glb,"\n");
   strcpy(str,"   B   I   N   G   O\n");
   putlog(glb,str);
   p = (int *) glb->card;
   j = 5;
   while (j--)
      {
      strcpy(str,"");
      i = 5;
      while (i--)
	 {
	 sprintf(cell,"%4d", *p);
	 strcat(str,cell);
	 p++;
	 } /* for each col in row */
      strcat(str,"\n");
      putlog(glb,str);
      } /* for each row */
   } /* putcard */

void initglb(glbfmt *glb)
   {
   int i;
   int ofst;
   int lastnum;
   int *p,*q;
   int *numptr;
   int *numend;
   int numlst[20];
   glb->run    = 1;
   glb->reset  = 0;
   glb->hit    = 0;
   glb->bingo  = 0;
   glb->wait   = 0;
   glb->drawnum = 999999;
   glb->prevnum = 999999;
   glb->loc[0] = '\0';
   glb->sd = (unsigned char *) rndinit();
   if (glb->sd == NULL)
      {
      fprintf(stderr,"bingo: out of memory "
	 "allocating glb->sd\n");
      exit(1);
      } /* if out of mem */
   /* unshuffle column 'B' */
   i = 1;
   p = (int *) numlst;
   q = (int *) numlst + 15;
   while (p < q)
      *p++ = i++;
   *p = 999999999;
   /* shuffle/fill column */
   lastnum = 15;
   p = (int *) glb->card;
   q = (int *) glb->card + 25;
   numend = (int *) numlst + 14;   /* 15 - 1 */
   while (p < q)
      {
      ofst = (int) rndnum(lastnum--,glb->sd);
      numptr = (int *) numlst + ofst;
      *p = *numptr;
      *numptr = *numend;
      *numend-- = 999999999;
      p += 5;
      } /* for each row in column 'B' */
   /* unshuffle column 'I' */
   i = 16;
   p = (int *) numlst;
   q = (int *) numlst + 15;
   while (p < q)
      *p++ =  i++;
   *p = 255;
   /* shuffle/fill column */
   lastnum = 15;
   p = (int *) glb->card + 1;
   q = (int *) glb->card + 25;
   numend = (int *) numlst + 14;   /* 15 - 1 */
   while (p < q)
      {
      ofst = (int) rndnum(lastnum--,glb->sd);
      numptr = (int *) numlst + ofst;
      *p = *numptr;
      *numptr = *numend;
      *numend-- = 999999999;
      p += 5;
      } /* for each row in column 'I' */
   /* unshuffle column 'N' */
   i = 31;
   p = (int *) numlst;
   q = (int *) numlst + 15;
   while (p < q)
      *p++ =  i++;
   *p = 255;
   /* shuffle/fill column */
   lastnum = 15;
   p = (int *) glb->card + 2;
   q = (int *) glb->card + 25;
   numend = (int *) numlst + 14;   /* 15 - 1 */
   while (p < q)
      {
      ofst = (int) rndnum(lastnum--,glb->sd);
      numptr = (int *) numlst + ofst;
      *p = *numptr;
      *numptr = *numend;
      *numend-- = 999999999;
      p += 5;
      } /* for each row in column 'N' */
   /* unshuffle column 'G' */
   i = 46;
   p = (int *) numlst;
   q = (int *) numlst + 15;
   while (p < q)
      *p++ =  i++;
   *p = 255;
   /* shuffle/fill column */
   lastnum = 15;
   p = (int *) glb->card + 3;
   q = (int *) glb->card + 25;
   numend = (int *) numlst + 14;   /* 15 - 1 */
   while (p < q)
      {
      ofst = (int) rndnum(lastnum--,glb->sd);
      numptr = (int *) numlst + ofst;
      *p = *numptr;
      *numptr = *numend;
      *numend-- = 999999999;
      p += 5;
      } /* for each row in column 'G' */
   /* unshuffle column 'O' */
   i = 61;
   p = (int *) numlst;
   q = (int *) numlst + 15;
   while (p < q)
      *p++ =  i++;
   *p = 255;
   /* shuffle/fill column */
   lastnum = 15;
   p = (int *) glb->card + 4;
   q = (int *) glb->card + 25;
   numend = (int *) numlst + 14;   /* 15 - 1 */
   while (p < q)
      {
      ofst = (int) rndnum(lastnum--,glb->sd);
      numptr = (int *) numlst + ofst;
      *p = *numptr;
      *numptr = *numend;
      *numend-- = 999999999;
      p += 5;
      } /* for each row in column 'O' */
   /* central square is blank */
   p = (int *) glb->card + 12;
   *p = 0;
   p = (int *) glb->cover;
   q = (int *) p + 25;
   while (p < q) *p++ = 0;
   opn(glb);
   putcard(glb);
   } /* initglb */

void initnc()
   {
   initscr();
   cbreak();
   noecho();
   nonl();
   intrflush(stdscr,FALSE);
   keypad(stdscr,TRUE);
   nodelay(stdscr,TRUE);
   curs_set(0);
   clear();
   refresh();
   } /* initnc */

void escape(glbfmt *glb)
   {
   int ch;
   ch = getch();
   if (0 && ch >= 0)
      {
      char str[64];
      move(20,0);
      sprintf(str,"%08x %d %u", ch, ch, ch);
      addstr(str);
      clrtoeol();
      refresh();
      } /* if char received */
   if (ch == 0x1b || ch == 'q')
      {
      glb->run = 0;
      } /* if quit */
   else if (ch == KEY_DOWN)
      {
      glb->hit = 1;
      } /* if cover */
   else if (ch == KEY_UP)
      {
      glb->bingo = 1;
      } /* if yell bingo */
   else if (ch == 'r')
      {
      glb->reset = 1;
      } /* if reset */
   } /* escape */

void bldmenu(glbfmt *glb)
   {
   move(0,30);
   addstr("Bingo");
   move(2,2);
   addstr("Down = Cover square");
   move(3,2);
   addstr("Up   = Bingo");
   move(4,2);
   addstr("r    = Reset");
   move(5,2);
   addstr("q    = Quit");
   refresh();
   } /* bldmenu */

void bldscr(glbfmt *glb)
   {
   int i,j;
   int row;
   int *p;
   int *cvr;
   char cell[16];
   char str[128];
   move(2,41);
   clrtoeol();
   addstr("    B    I    N    G    O");
   p = (int *) glb->card;
   cvr = (int *) glb->cover;
   row = 4;

   j = 5;
   while (j--)
      {
      strcpy(str,"");
      i = 5;
      while (i--)
         {
         if (*cvr) strcpy(cell,"   **");
         else sprintf(cell,"%5d", *p);
         strcat(str,cell);
         cvr++;
         p++;
         } /* for each number in row */
      move(row,41);
      clrtoeol();
      addstr(str);
      row += 2;
      } /* for each row on card */

   refresh();
   } /* bldscr */

void coverfailedmsg(glbfmt *glb)
   {
   char str[128];
   void slp(glbfmt *glb);
   move(20,25);
   clrtoeol();
   strcpy(str,"Cover failed");
   addstr(str);
   refresh();
   while (1)
      {
      /* slp(glb); */
      escape(glb);
      if (!glb->run || glb->reset) break;
      sleep(1);
      } /* wait for quit */
   } /* coverfailedmsg */

void cover(glbfmt *glb)
   {
   int coversw;
   int dblcoversw;
   int *p,*q;
   int *cvr;
   char str[64];
   dblcoversw = coversw = 0;
   p = (int *) glb->card;
   q = (int *) p + 25;
   cvr = (int *) glb->cover;
   while (p < q)
      {
      if (glb->prevnum == *p)
	 {
	 *cvr = 1;
         sprintf(str,"Cover previous %2d ",
	    glb->prevnum);
         putlog(glb,str);
	 glb->prevnum += 1000;
	 coversw = 1;
	 break;
	 } /* match previous number */
      cvr++;
      p++;
      } /* for each square on card */
   p = (int *) glb->card;
   q = (int *) p + 25;
   cvr = (int *) glb->cover;
   while (!coversw && p < q)
      {
      if (glb->drawnum == *p)
	 {
	 *cvr = 1;
         sprintf(str,"Cover %2d ", glb->drawnum);
         putlog(glb,str);
	 glb->drawnum += 1000;
	 coversw = 1;
	 break;
	 } /* match current number */
      cvr++;
      p++;
      } /* for each square on card */
   if (glb->drawnum > 1000
      && glb->drawnum <= 1075)
      dblcoversw = 1;
   if (!coversw && !dblcoversw)
      {
      if (glb->drawnum > 1000
	 && glb->drawnum <= 1075)
         glb->drawnum -= 1000;
      if (glb->prevnum > 1000
	 && glb->prevnum <= 1075)
         glb->prevnum -= 1000;
      if (glb->prevnum <= 75)
	 {
         sprintf(glb->loc,"\nEither %2d "
	    "or %2d not matched on card",
	    glb->prevnum, glb->drawnum); 
	 } /* if not first draw */
      else
	 {
         sprintf(glb->loc,"\nDrawn number %2d "
	    "not matched on card",
	    glb->drawnum); 
	 } /* if first draw */
      coverfailedmsg(glb);
      } /* invalid cover */
   } /* cover */

void bingomsg(glbfmt *glb)
   {
   void slp(glbfmt *glb);
   move(20,23);
   clrtoeol();
   addstr(glb->loc);
   refresh();
   while (1)
      {
      /* slp(glb); */
      escape(glb);
      if (!glb->run || glb->reset) break;
      sleep(1);
      } /* wait for quit */
   } /* bingomsg */

void bingofailedmsg(glbfmt *glb)
   {
   char str[128];
   void slp(glbfmt *glb);
   move(20,22);
   clrtoeol();
   strcpy(str,"Bingo match failed");
   addstr(str);
   refresh();
   while (1)
      {
      /* slp(glb); */
      escape(glb);
      if (!glb->run || glb->reset) break;
      sleep(1);
      } /* wait for quit */
   } /* bingofailedmsg */

void bingo(glbfmt *glb)
   {
   int totmtch;
   int *p,*q;
   int *cvr;
   char str[128];
   glb->bingo += 2;
   if (glb->drawnum > 1000)
      glb->drawnum -= 1000;
   if (glb->prevnum > 1000)
      glb->prevnum -= 1000;
   p = (int *) glb->card;
   q = (int *) p + 25;
   cvr = (int *) glb->cover;
   while (p < q)
      {
      if (glb->prevnum == *p)
	 {
	 *cvr = 1;
	 bldscr(glb);
         sprintf(str,"Cover previous %2d "
	    "during bingo\n",
            glb->prevnum);
         putlog(glb,str);
	 break;
	 } /* match previous number */
      cvr++;
      p++;
      } /* for each square in card */
   p = (int *) glb->card;
   q = (int *) p + 25;
   cvr = (int *) glb->cover;
   while (p < q)
      {
      if (glb->drawnum == *p)
	 {
	 *cvr = 1;
	 bldscr(glb);
         sprintf(str,"Cover %2d during bingo\n",
            glb->drawnum);
         putlog(glb,str);
	 break;
	 } /* match current number */
      cvr++;
      p++;
      } /* for each square in card */
   /* test for bingo in row 1 */
   totmtch = 0;
   p = (int *) glb->cover;
   q = (int *) p + 5;
   while (p < q)
      {
      if (*p) totmtch++;
      p++;
      } /* for each square in row */
   if (totmtch == 5)
      {
      strcpy(glb->loc,"Bingo in row 1!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in row 2 */
   totmtch = 0;
   p = (int *) glb->cover + 5;
   q = (int *) p + 5;
   while (p < q)
      {
      if (*p) totmtch++;
      p++;
      } /* for each square in row */
   if (totmtch == 5)
      {
      strcpy(glb->loc,"Bingo in row 2!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in row 3 */
   totmtch = 0;
   p = (int *) glb->cover + 10;
   q = (int *) p + 5;
   while (p < q)
      {
      if (*p) totmtch++;
      p++;
      } /* for each square in row */
   if (totmtch == 4)
      {
      strcpy(glb->loc,"Bingo in row 3!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in row 4 */
   totmtch = 0;
   p = (int *) glb->cover + 15;
   q = (int *) p + 5;
   while (p < q)
      {
      if (*p) totmtch++;
      p++;
      } /* for each square in row */
   if (totmtch == 5)
      {
      strcpy(glb->loc,"Bingo in row 4!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in row 5 */
   totmtch = 0;
   p = (int *) glb->cover + 20;
   q = (int *) p + 5;
   while (p < q)
      {
      if (*p) totmtch++;
      p++;
      } /* for each square in row */
   if (totmtch == 5)
      {
      strcpy(glb->loc,"Bingo in row 5!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in col 'B' */
   totmtch = 0;
   p = (int *) glb->cover;
   q = (int *) p + 25;
   while (p < q)
      {
      if (*p) totmtch++;
      p += 5;
      } /* for each square in col */
   if (totmtch == 5)
      {
      strcpy(glb->loc,"Bingo in column 'B'!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in col 'I' */
   totmtch = 0;
   p = (int *) glb->cover + 1;
   q = (int *) p + 25;
   while (p < q)
      {
      if (*p) totmtch++;
      p += 5;
      } /* for each square in col */
   if (totmtch == 5)
      {
      strcpy(glb->loc,"Bingo in column 'I'!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in col 'N' */
   totmtch = 0;
   p = (int *) glb->cover + 2;
   q = (int *) p + 25;
   while (p < q)
      {
      if (*p) totmtch++;
      p += 5;
      } /* for each square in col */
   if (totmtch == 4)
      {
      strcpy(glb->loc,"Bingo in column 'N'!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in col 'G' */
   totmtch = 0;
   p = (int *) glb->cover + 3;
   q = (int *) p + 25;
   while (p < q)
      {
      if (*p) totmtch++;
      p += 5;
      } /* for each square in col */
   if (totmtch == 5)
      {
      strcpy(glb->loc,"Bingo in column 'G'!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in col 'O' */
   totmtch = 0;
   p = (int *) glb->cover + 4;
   q = (int *) p + 25;
   while (p < q)
      {
      if (*p) totmtch++;
      p += 5;
      } /* for each square in col */
   if (totmtch == 5)
      {
      strcpy(glb->loc,"Bingo in column 'O'!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in NW-SE */
   totmtch = 0;
   p = (int *) glb->cover;
   q = (int *) p + 25;
   while (p < q)
      {
      if (*p) totmtch++;
      p += 6;
      } /* for each square in col */
   if (totmtch == 4)
      {
      strcpy(glb->loc,"Bingo in diagonal NW-SE!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   /* test for bingo in NE-SW */
   totmtch = 0;
   p = (int *) glb->cover + 4;
   q = (int *) p + 21;
   while (p < q)
      {
      if (*p) totmtch++;
      p += 4;
      } /* for each square in col */
   if (totmtch == 4)
      {
      strcpy(glb->loc,"Bingo in diagonal NE-SW!");
      bingomsg(glb);
      return;
      } /* valid bingo */
   sprintf(glb->loc,"\nBingo match failed "
      "on number %d", glb->drawnum);
   bingofailedmsg(glb);
   } /* bingo */

void slp(glbfmt *glb)
   {
   int rslt;
   time_t strt;
   time_t lmt;
   time_t now;
   char str[32];
   struct timespec request;
   struct timespec remain;
   time(&strt);
   lmt = strt + glb->wait;
   now = strt;
   request.tv_sec  = 0;
   request.tv_nsec = 1;
   remain.tv_sec   = 0;
   remain.tv_nsec  = 0;
   while (now < lmt)
      {
      rslt = nanosleep(&request,&remain);
      if (rslt < 0)
	 {
	 if (errno == EINTR)
	    strcpy(str,"EINTR");
	 else if (errno == EINVAL)
	    strcpy(str,"EINVAL");
	 else
	    strcpy(str,"unknown value");
	 fprintf(stderr,"req sec %u "
	    " req nsec %u "
	    " rem sec %u "
	    " rem nsec %u\n",
            (unsigned int) request.tv_sec,
            (unsigned int) request.tv_nsec,
	    (unsigned int) remain.tv_sec,
	    (unsigned int) remain.tv_nsec);
	 fprintf(stderr,"bingo: slp: error %d %s\n",
	    errno, str);
	 perror("Nanosleep error");
	 exit(1);
	 } /* if err */
      escape(glb);
      if (!glb->run || glb->reset) break;
      if (glb->hit)
	 {
	 cover(glb);
	 bldscr(glb);
	 glb->hit = 0;
	 } /* if hit */
      else if (glb->bingo && glb->run)
	 {
	 bingo(glb);
	 glb->bingo++;
	 return;
	 } /* if bingo */
      time(&now);
      } /* wait loop */
   } /* slp */

void play(glbfmt *glb)
   {
   int i;
   int ofst;
   int lastnum;
   int *p,*q;
   int *numptr;
   int *numend;
   int numlst[75];
   char logmsg[64];
   char scrmsg[64];
   char str[64];
   p = (int *) numlst;
   q = (int *) numlst + 75;
   i = 1;
   while (p < q) *p++ = i++;
   numend = (int *) numlst + 74;    /* 75 - 1 */
   lastnum = 75;
   while (glb->run && lastnum >= 0)
      {
      ofst = (int) rndnum(lastnum--,glb->sd);
      numptr = (int *) numlst + ofst;
      glb->prevnum = glb->drawnum;
      glb->drawnum = *numptr;
      *numptr = *numend;
      *numptr = *numend;
      *numend-- = 999999999;
      move(3,30);
      if (glb->drawnum <= 15)
         strcpy(str,"B");
      else if (glb->drawnum <= 30)
         strcpy(str,"I");
      else if (glb->drawnum <= 45)
         strcpy(str,"N");
      else if (glb->drawnum <= 60)
         strcpy(str,"G");
      else
         strcpy(str,"O");
      sprintf(scrmsg,"%s %d  ",
	 str, glb->drawnum);
      addstr(scrmsg);
      refresh();
      sprintf(logmsg,"\nDraw "
         "%s %2d ",
	 str, glb->drawnum);
      putlog(glb,logmsg);
      escape(glb);
      if (!glb->run || glb->reset) return;
      slp(glb);
      if (!glb->run || glb->reset) return;
      if (glb->hit)
	 {
	 cover(glb);
	 bldscr(glb);
	 glb->hit = 0;
	 } /* if cover square */
      else if (glb->bingo == 1)
	 {
	 bingo(glb);
	 } /* if bingo == 1 */
      } /* for each draw */
   } /* play */

void run(glbfmt *glb)
   {
   int wait;
   while (glb->run)
      {
      bldscr(glb);
      escape(glb);
      if (!glb->run) break;
      if (glb->run && glb->reset)
	 {
	 wait = glb->wait;
	 cls(glb);
	 clear();
	 bldmenu(glb);
	 initglb(glb);
	 /* glb->reset is set to zero */
         bldscr(glb);
	 glb->wait = wait;
	 play(glb);
	 } /* if reset */
      else
	 {
	 play(glb);
	 } /* not reset */
      if (glb->run && !glb->reset) slp(glb);
      } /* display iteration */
   } /* run */

int main(int argc,char **argv)
   {
   int wait;
   char str[128];
   glbfmt *glb;
   if (argc != 2) putstx(*argv);
   wait = atoi(*(argv+1));
   if (wait < 1 || wait > 10)
      {
      fprintf(stderr,"Invalid interval\n");
      putstx(*argv);
      } /* invalid interval */
   glb = (glbfmt *) malloc(sizeof(glbfmt));
   if (glb == NULL)
      {
      fprintf(stderr,"bingo: out of memory "
	 "allocating glb\n");
      exit(1);
      } /* if out of mem */
   initglb(glb);
   glb->wait = wait;
   initnc();
   bldmenu(glb);
   bldscr(glb);
   run(glb);
   endwin();
   sprintf(str,"%s\n", glb->loc);
   putlog(glb,str);
   printf("%s\n", glb->loc);
   cls(glb);
   free(glb->sd);
   free(glb);
   return(0);
   } /* main */
